--==================================================================================================--
--	chataclysm (noun)
--	1) a disconnect that results in a user losing all their chat channels 
--	2) the name of an addon which automatically repairs the damage caused by a chataclysmic event
--	source: The Medivh-Wrynn Common Dictionary (Revised Edition)
--==================================================================================================--

Chataclysm = LibStub("AceAddon-3.0"):NewAddon("Chataclysm", "AceConsole-3.0", "AceEvent-3.0", "AceTimer-3.0")


--==== set upvalues ====--
-- GLOBALS: LibStub 
local _G, Chataclysm, NUM_CHAT_WINDOWS, guiHidden = _G, Chataclysm, NUM_CHAT_WINDOWS, guiHidden
local date, format, math, mod, pairs, select, strfind, tinsert, tonumber, tostring, tremove, type = date, format, math, mod, pairs, select, strfind, tinsert, tonumber, tostring, tremove, type
local ChatFrame_AddChannel, ChatFrame_RemoveChannel, CollapseChannelHeader, ExpandChannelHeader, GetChannelDisplayInfo, GetChannelName, GetChatWindowChannels, GetNumDisplayChannels, GetTime, JoinPermanentChannel, JoinTemporaryChannel, LeaveChannelByName = ChatFrame_AddChannel, ChatFrame_RemoveChannel, CollapseChannelHeader, ExpandChannelHeader, GetChannelDisplayInfo, GetChannelName, GetChatWindowChannels, GetNumDisplayChannels, GetTime, JoinPermanentChannel, JoinTemporaryChannel, LeaveChannelByName
--======================--


-- Chat Actions
local CA = {
	"INIT",
	"DONE",
	"ADD",
	"REMOVE",
	"BUFFER",
	"ADD_BUFFER",
	"REMOVE_BUFFER",
	"PLACEHOLDER",
	"ADD_PLACEHOLDER",
	"REMOVE_PLACEHOLDER",
}
local g_actions = {} -- key(int) = id of channel, value(table) = { "type" = action scheduled to happen to channel (CA), "name" = channel we want to see in this slot }

local DL = {
	[0] = "none",
	"standard",
	"error",
	"test",
	"config",
	"events",
	"spam",
}
local dbgprint = { squelch = function(...) return nil end, 
                   print   = function(...) return Chataclysm:Print(...) end }
local function SetDebugLevel(iDL)
	for i = 1,     iDL do dbgprint[ DL[i] ] = dbgprint.print; end
	for i = iDL+1, #DL do dbgprint[ DL[i] ] = dbgprint.squelch; end
end

local Event = { module = Chataclysm }
Event.__index = Event
function Event:Release()
	if self.handler then 
		self.module:UnregisterEvent(self.eventName)
		self.handler = nil
	end
end
function Event:Set(handler)
	if handler == self.handler then return end
	self:Release()
	self.module:RegisterEvent(self.eventName, handler)
	self.handler = handler
end

local Event_ChannelNotice = setmetatable( { eventName = "CHAT_MSG_CHANNEL_NOTICE" }, Event )
local Event_ChannelNoticeUser = setmetatable( { eventName = "CHAT_MSG_CHANNEL_NOTICE_USER" }, Event )
local Event_ZoneChanged = setmetatable( { eventName = "ZONE_CHANGED_NEW_AREA" }, Event )


local g_strTempChannelPrefix = "ChataclysmPlaceholder"
local g_strTempChannelFormat = ""
local function IsTempChannel(name) return strfind(name, g_strTempChannelPrefix, 1, true) end

local g_dbFormat = 1
local g_defaults = {
	global = {
		debugLevel = 2,        -- DL setting for verbosity of debug output 
		initialDelay = 10,     -- number of seconds until we will check/fix after addon load unless we see a CHANNEL_UI_UPDATE event 
		incrementalDelay = 2,  -- number of seconds until we will check/fix after seeing a CHANNEL_UI_UPDATE event unless we see another CHANNEL_UI_UPDATE event
	},
	profile = {
		backup = {
			channels = {}, -- key(int) = channel number, value(string) = channel name
			frames   = {}, -- key(string) = channel name , value(table) = { key(string) name of chat frame, value = true }
		},
		saved = {
			channels = {}, -- key(int) = channel number, value(string) = channel name
			frames   = {}, -- key(string) = channel name , value(table) = { key(string) name of chat frame, value = true }
		},
	},
}

function Chataclysm:ValidateDBFormat()
	if (g_dbFormat == self.db.profile.dbFormat) then
		return
	end

	-- blow away settings from Beta 1 version
	if (not self.db.profile.dbFormat) then
		self.db:ResetProfile()
		self.db.global.formats = nil
		
		self.db.profile.dbFormat = 1
	end
end

-- Called when the addon is loaded
function Chataclysm:OnInitialize()
	self.db = LibStub("AceDB-3.0"):New("ChataclysmDB", g_defaults)
	self.db.RegisterCallback(self, "OnDatabaseShutdown", "SaveCurrentChatConfig")
	self:ValidateDBFormat()
	
	LibStub("AceConfig-3.0"):RegisterOptionsTable("Chataclysm", Chataclysm.GetUIOptions, { "Chataclysm" } )
	
	for k,v in pairs(CA) do CA[v] = k end
	for k,v in pairs(DL) do DL[v] = k end
	SetDebugLevel(self.db.global.debugLevel)
	
	local seed = math.modf(mod((GetTime() * 1000), 100000))
	g_strTempChannelFormat = format("%s%%i%i", g_strTempChannelPrefix, (seed))
	dbgprint.spam(g_strTempChannelFormat)

	-- init actions struct
	for i=1, 10 do
		g_actions[i] = { type = CA.INIT, name = self.db.profile.saved.channels[i] }
	end
end

-- Called when the addon is enabled
function Chataclysm:OnEnable()
	Event_ChannelNotice:Set("DelayFirstChatCheck")
	self.timer = self:ScheduleTimer("FirstChatCheck", self.db.global.initialDelay)
end

-- Called when the addon is Disabled
function Chataclysm:OnDisable()
	Event_ChannelNotice:Release()
	Event_ChannelNoticeUser:Release()
	Event_ZoneChanged:Release()
end

-- returns { key(string) = name of chat frame, value(table) = { "frame" = the chatframe itself (table), "hasChannel" = table { key = channel name (string), value = true (if exists) } } }
local function GetCurrentFrames()
	local frames = {}

	local function GetChannelList(...)
		local tbl = {}
		for iOffset = 1, select("#", ...), 2 do
			tbl[select(iOffset, ...)] = true
		end
		return tbl;
	end

	for i=1, NUM_CHAT_WINDOWS do
		local tblChatFrame = _G["ChatFrame" .. tostring(i)]
		if (tblChatFrame ~= nil) then
			local channels = GetChannelList(GetChatWindowChannels(tblChatFrame:GetID()))
			frames[tblChatFrame.name] = { frame = tblChatFrame, hasChannel = channels }
		end
	end

	for k,v in pairs(frames) do
		dbgprint.spam("frames", k, v.frame:GetID())
		for n,_ in pairs(v.hasChannel) do
			dbgprint.spam("  ", n)
		end
	end
	
	return frames
end

-- returns { key(int) = id of channel, value(str) = current name (or server name) of the channel (generalized, no zone) }
-- also supports reverse look up { key(str) name of channel, value(int) id of channel }
local function GetCurrentChannels()
	local tblChannels = {}
	local tblCollapsedHeaders = {}
	local iChannels = GetNumDisplayChannels()
	local i = 1
	
	while (i <= iChannels) do
		local name, header, collapsed, channelNumber, count, active, category = GetChannelDisplayInfo(i)
		if (header) then
			if (collapsed and count) then
				tinsert(tblCollapsedHeaders, i)
				ExpandChannelHeader(i)
				iChannels = GetNumDisplayChannels()
			end
		else
			dbgprint.spam("display channel", channelNumber, name)
			tblChannels[channelNumber] = name
			tblChannels[name] = channelNumber
		end
		i = i + 1
	end
	
	while #tblCollapsedHeaders > 0 do
		local channelNumber = tremove(tblCollapsedHeaders)
		CollapseChannelHeader(channelNumber)
	end
	
	return tblChannels
end

function Chataclysm:EventHandler(...) dbgprint.events(...) end

function Chataclysm:DelayFirstChatCheck(event, message, _, _, _, _, _, _, channelNumber, channelName, ...)
	dbgprint.events("DelayFirstChatCheck", event, message, channelNumber, channelName)
	if self.timer then
		self:CancelTimer(self.timer, false)
		self.timer = self:ScheduleTimer("FirstChatCheck", self.db.global.incrementalDelay)
	end
end

function Chataclysm:FirstChatCheck()
	dbgprint.config("Chataclysm is now checking your chat channels ...")
	self.timer = nil

	Event_ChannelNotice:Set("ResolveActionNotice")
	Event_ChannelNoticeUser:Set("ResolveActionNoticeUser")

	self:CheckChat()
end

function Chataclysm:RetryPlaceholders(...)
	dbgprint.events("RetryPlaceholders", ...)
	for iChannel = 1, 10 do
		local action = g_actions[iChannel] or {}
		if (action.type == CA.PLACEHOLDER) then
			action.type = CA.REMOVE_PLACEHOLDER
		end
	end

	self:ScheduleTimer("CheckChat", 5)
end


function Chataclysm:ResolveActionNoticeUser(event, message, _, _, _, _, _, zoneChannel, channelNumber, channelName, ...)
	dbgprint.events("ResolveActionNoticeUser", event, message, zoneChannel, channelNumber, channelName)
	local action = g_actions[channelNumber]
	if (message == "INVALID_NAME" and action.type == CA.ADD and zoneChannel ~= 0) then
		dbgprint.standard("Chataclysm is creating a placeholder channel for", channelNumber, channelName, "until conditions allow us to rejoin")
		action.type = CA.ADD_PLACEHOLDER
		
		Event_ZoneChanged:Set("RetryPlaceholders")
		self:ScheduleTimer("CheckChat", 0.3)
	end
end

function Chataclysm:ResolveActionNotice(event, message, _, _, _, _, _, _, channelNumber, channelName, ...)
	dbgprint.events("ResolveActionNotice", event, message, channelNumber, channelName)
	local action = g_actions[channelNumber]

	if (message == "YOU_LEFT" and action.type ~= CA.DONE) then
		if (action.type == CA.REMOVE or action.type == CA.REMOVE_PLACEHOLDER) then
			action.type = CA.INIT
		elseif (action.type == CA.REMOVE_BUFFER) then
			action.type = CA.DONE
		else
			dbgprint.error("left channel for unexpected reason", channelNumber, channelName)
		end
		self:ScheduleTimer("CheckChat", 0.3)

	elseif (message == "YOU_JOINED") then
		local channels = GetCurrentChannels()
		if (action.type == CA.DONE) then
			-- they may be rejoining a channel we know something about
			self:EnsureChatWindows(channels[channelNumber])
			return
		end
		
		if (action.type == CA.ADD and channelNumber == channels[action.name]) then
			action.type = CA.DONE
		elseif (action.type == CA.ADD_BUFFER and IsTempChannel(channelName)) then
			action.type = CA.BUFFER
		elseif (action.type == CA.ADD_PLACEHOLDER and IsTempChannel(channelName)) then
			action.type = CA.PLACEHOLDER
		else
			dbgprint.error("ERROR: channel joined but name does not match expected", action.name, GetChannelName(action.name))
			action.type = CA.DONE -- using CA.INIT is likely to create loops
		end
		self:CheckChat()
	end
end

function Chataclysm:CheckChat()
	self:UpdateActionTable()
	if (self:FExecuteNextAction()) then 
		return 
	end
	
	-- if we had no joins or leaves then we can remove any BUFFER channels added for padding
	dbgprint.spam("removing any buffers")
	for iSaved=1, 10 do
		if (g_actions[iSaved].type == CA.BUFFER) then 
			g_actions[iSaved].type = CA.REMOVE_BUFFER
		end
	end
	if (self:FExecuteNextAction()) then
		return
	end
	
	dbgprint.standard("Chataclysm has finished checking your channels")
	
	-- if there are no placeholder channels then we've finished the task of checking channels and can stop monitoring events
	for iSaved=1, 10 do
		if (g_actions[iSaved].type == CA.PLACEHOLDER) then 
			return
		end
	end

	dbgprint.spam("removing event handlers")
	Event_ChannelNotice:Release()
	Event_ChannelNoticeUser:Release()
	Event_ZoneChanged:Release()
end


function Chataclysm:EnsureChatWindows(channelName, currentFrames)
	if (not channelName) then return end
	currentFrames = currentFrames or GetCurrentFrames()
	
	local savedFrame = self.db.profile.saved.frames[channelName]
	if (type(savedFrame) ~= "table") then
		dbgprint.config(channelName, "apparently was not displayed in any chat windows")
	else	
		for savedFrameName,_ in pairs(self.db.profile.saved.frames[channelName]) do
			local frame = currentFrames[savedFrameName]
			if (frame == nil) then
				dbgprint.error(channelName, "cannot be added to chat window", savedFrameName, "as it no longer exists")
			elseif (not frame.hasChannel[channelName]) then
				dbgprint.config("adding channel", channelName, "to chat window", savedFrameName)
				-- If the channel is joined to the channel but doesn't exactly match case, then messages
				-- will still show in the window but the checkbox will be unchecked. An attempt to add a
				-- channel a second time with a different case will fail because it sees it as the same
				-- channel. For that same reason though, a remove works even with different casing. 
				-- So ... before adding, first do a remove to clear a wrong cased version of the channel.
				ChatFrame_RemoveChannel(frame.frame, channelName)
				ChatFrame_AddChannel(frame.frame, channelName)
			end
		end
	end
end


-- returns true if it made a call to add or remove a channel
-- when adding a channel it also repairs associations with chat windows
function Chataclysm:FExecuteNextAction()
	for iChannel=1, 10 do
		local action = g_actions[iChannel].type
		if (action == CA.ADD) then
			local strChannelName = g_actions[iChannel].name
			JoinPermanentChannel(strChannelName)
			self:EnsureChatWindows(strChannelName)
			return true;

		elseif (action == CA.ADD_BUFFER or 
		        action == CA.ADD_PLACEHOLDER ) then
			JoinTemporaryChannel(format(g_strTempChannelFormat, iChannel))
			return true;

		elseif (action == CA.REMOVE or 
		        action == CA.REMOVE_BUFFER or  
		        action == CA.REMOVE_PLACEHOLDER ) then
			LeaveChannelByName(select(2, GetChannelName(iChannel)))
			return true;
		end
	end		
end
		

function Chataclysm:UpdateActionTable()
	local channels = GetCurrentChannels()
	local frames = GetCurrentFrames()

	local function FEnsureSlotting(index)
		-- recursively ensure that there are channels below the slot we want to fill, 
		-- filling them with a temp channel if need be
		index = index - 1
		
		if (index == 0) then 
			return true;
		end
		
		-- if there are remove operations scheduled we should wait for those to finish 
		-- processing before trying to add new channels to slots after them
		local action = g_actions[index].type
		if (action == CA.REMOVE or 
			action == CA.REMOVE_BUFFER or 
			action == CA.REMOVE_PLACEHOLDER ) then 
			return false;
		end
		
		if (action == CA.DONE and not channels[index]) then
			g_actions[index].type = CA.ADD_BUFFER
		end
		
		return FEnsureSlotting(index)
	end


	for iSaved=1, 10 do
		-- we only need evaluate channels that are still in the INIT action state
		if (g_actions[iSaved].type == CA.INIT) then
			local strSaved = g_actions[iSaved].name
			if (strSaved == nil or strSaved == channels[iSaved]) then
				-- no saved channel for this index or we'ved already joined the channel
				g_actions[iSaved].type = CA.DONE
				self:EnsureChatWindows(strSaved, frames) -- blizz sometimes remembers server channels but not their windows
			else
				-- are we in the wrong slot? (as opposed to just absent)
				local iActual = channels[strSaved]
				if (iActual ~= nil) then
					dbgprint.config(strSaved, "was found in slot", iActual, "and needs to be removed first so that it can be added to slot", iSaved)
					g_actions[iActual].type = CA.REMOVE
					return
				end				

				-- is there another channel sitting in our spot?
				if (channels[iSaved] ~= nil) then
					dbgprint.config(strSaved, "cannot be added until", channels[iSaved], "is removed from its slot")
					-- TODO REVIEW should we save removed channels and append after fixing?
					g_actions[iSaved].type = CA.REMOVE
					return
				end

				-- not safe to add if there are removals in the works
				if (not FEnsureSlotting(iSaved)) then
					return
				end

				dbgprint.error("fix needed for", iSaved, strSaved)
				g_actions[iSaved].type = CA.ADD
			end
		end
	end
end


function Chataclysm:SaveChatConfig(strBranch)
	-- self.db.profile[strBranch] = self.db.profile[strBranch] or {}

	local channels = {}
	local currentChannels = GetCurrentChannels()
	for i=1, 10 do
		local action = g_actions[i]
		channels[i] = currentChannels[i]
		
		-- if we're seeing a channel marked INIT then we probably are in the process of dc'ing. save nothing!
		if (action.type == CA.INIT) then
			return;

		-- if we haven't been able to replace a placeholder channel yet, then make sure
		-- the channel we were holding a spot for is saved out for next session
		elseif (action.type == CA.PLACEHOLDER) then
			dbgprint.config("placeholder channel flag still set for", i, channels[i])
			if (IsTempChannel(channels[i])) then
				channels[i] = action.name
			end

		-- if a buffer channel is somehow still there, then just nuke it
		elseif (action.type == CA.BUFFER) then
			dbgprint.error("buffer channel flag still set for", i, channels[i])
			if (IsTempChannel(channels[i])) then
				channels[i] = nil
			end
		end
	end
	self.db.profile[strBranch].channels = channels


	-- We're choosing not to throw away channel-frame associations for channels we're no
	-- longer joined to. Should you rejoin a channel later, having it filtered correctly 
	-- automatically seems a good user experience. So overwrite any existing channels 
	-- subtable, but don't just replace the entire frameList
	--	local savedFrames = self.db.profile[strBranch].frames
	-- 
	-- Actually, scrap that for now and overwrite. channel name mismatches can cause wonky behavior.
	-- maybe do it again if we go with squash it all to lower for save using format for rehydration.
	local frames = {}
	local currentFrames = GetCurrentFrames()
	for frameName,frame in pairs(currentFrames) do
		for channelName,_ in pairs(frame.hasChannel) do
			if (not frames[channelName]) then frames[channelName] = {} end
			frames[channelName][frameName] = true
		end
	end
	self.db.profile[strBranch].frames = frames

	self.db.profile[strBranch].lastSave = date()
end

function Chataclysm:SaveCurrentChatConfig()
	self:SaveChatConfig("saved")
end

local function TableCopy(source, dest)
	local new = dest or {}
	for k,v in pairs(source) do
		if (type(k) == "table") then
			dbgprint.error("ERROR - TableCopy behavior when keys are of type table is undefined")
		end
		if (type(v) == "table") then
			new[k] = TableCopy(v)
		else
			new[k] = v
		end
	end
	return new
end


function Chataclysm:Debug(info, value)
	local strValue, index = self:GetArgs(value, 1);

	if (strValue ~= nil) then
		if (index ~= 1e9) then
			dbgprint.standard("ERROR: too many parameters passed");
		end

		local iValue = tonumber(strValue)
		if (iValue == nil) then
			dbgprint.standard("ERROR: no value given for debug level");
			return;
		elseif (DL[iValue] == nil) then
			dbgprint.standard("ERROR: level_value must be number between 0 and", #DL);
			return;
		else
			self.db.global.debugLevel = iValue
			SetDebugLevel(iValue)
			dbgprint.standard("debug level set to", DL[iValue])
		end
	end
end


--=========================================================================--
-- Options UI definition
--=========================================================================--

function Chataclysm.GetUIOptions()
	if (not Chataclysm.options) then
		Chataclysm.options = {
			name = "Chataclysm",
			handler = Chataclysm,
			type = 'group',
			args = {
--[=====[
do
	local realstrlower = strlower
	local strlower = function(str)
		if (str == nil) then return nil end
		return realstrlower(str)
	end
end

				format = {
					hidden = guiHidden,
					order = 5,
					type = "input",
					name = "format",
					desc = "sets the form of the channel name to use when creating the channel",
					get = function(info) return "" end,
					set = function(info, value)
						local strKey, index = Chataclysm:GetArgs(value, 1, 1)
						if (strKey == nil) then
							local channels = {}
							for _,v in pairs(Chataclysm.db.global.formats) do
								tinsert(channels, v)
							end
							table.sort(channels)
							for _,v in ipairs(channels) do
								dbgprint.standard(v)
							end
						else
							repeat 
								Chataclysm.db.global.formats[strlower(strKey)] = strKey
								strKey, index = Chataclysm:GetArgs(value, 1, index);
							until index == 1e9
						end
					end,
				},
--]=====]
				debug = {
					hidden = true,
					order = 0,
					type = "input",
					name = "debug",
					desc = "Sets verbosity of debug output",
					get = function(info) return "" end,
					set = "Debug",
				},
				leaveall = {
					hidden = true,
					order = 0,
					type = "execute",
					name = "leaveall",
					desc = "debug command - leave all channels",
					func = function(info)
						for i=1, 10 do
							local name = select(2, GetChannelName(i))
							if (name ~= nil) then LeaveChannelByName(name) end
						end
					end,
				},
				backup = {
					hidden = guiHidden,
					order = 3,
					type = "execute",
					name = "backup",
					desc = "create a save point for use with the restore command",
					func = function(info) Chataclysm:SaveChatConfig("backup") end,
				},
				restore = {
					hidden = guiHidden,
					order = 4,
					type = "execute",
					name = "restore",
					desc = "sets chat settings to state saved with backup command",
					func = function(info)
						-- if called before we automatically kick in our normal first check, 
						-- canel the timers so that FirstChatCheck doesn't get called twice
						Chataclysm:CancelAllTimers()
						
						-- blast over the current profile with the backup config
						Chataclysm.db.profile.saved = TableCopy(Chataclysm.db.profile.backup)

						-- init actions struct
						for i=1, 10 do
							g_actions[i] = { type = CA.INIT, name = Chataclysm.db.profile.saved.channels[i] }
						end
						
						Chataclysm:FirstChatCheck()
					end,
				},
			},
		}
	end

	return Chataclysm.options;
end

